<?php

include('version.php');
require_once('Owncloud/webdav.php');
require_once('Owncloud/authexception.php');
require_once('Owncloud/cachemanager.php');

/**
 * Owncloud Module
 */
class OwncloudListModule extends ListModule {

	private $wdc;				// webdav client
	private $cache;				// cache manager
	private $initdone = false;

	/**
	 * Constructor
	 * @param int $id unique id.
	 * @param array $data list of all actions.
	 */
	function OwncloudListModule($id, $data) {
		parent::ListModule($id, $data);

		$this->start = 0;
	}

	/**
	 * Creates the notifiers for this module,
	 * and register them to the Bus.
	 */
	/*function createNotifiers() {
		$entryid = $this->getEntryID();
		$GLOBALS["bus"]->registerNotifier('owncloudlistnotifier', $entryid);
	}*/

/**
	 * Initialize PHP API of Owncloud
	 * @param $data
	 */
	private function bootstrap($data) {
		$this->wdc = new webdav_client();
		$oc_base = $GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/owncloud_path");
		$oc_server = $GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/server");
		$oc_ssl = $GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/use_ssl");
		
		$this->wdc->set_server($oc_server);
		if($oc_ssl == true) {
			$this->wdc->set_port(intval($GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/port_ssl")));
		} else {
			$this->wdc->set_port(intval($GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/port")));
		}
		$this->wdc->set_ssl($oc_ssl);
		$this->wdc->set_pathprefix($oc_base);
		// use HTTP/1.1
		$this->wdc->set_protocol(1);
		// enable debugging
		$this->wdc->set_debug($_ENV["OC_PLUGIN_DEBUG"]);
		
		// initialize the cachemanager
		$this->cache = new cache_manager(PLUGIN_OWNCLOUDBROWSER_TMP, $GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/enable_caching"));
	}

	/**
	 * Executes all the actions in the $data variable.
	 * Exception part is used for authentication errors also
	 * @return boolean true on success or false on failure.
	 */
	public function execute() {
		$result = false;

		foreach($this->data as $actionType => $actionData)
		{			
			/* we have to fix this.... */
			//TODO: rewrite this dirty code...
			if(!$this->initdone && isset($actionData)) {
				$this->bootstrap($actionData);
				$this->initdone = true;
			}
			if(isset($actionType)) {
				try {
					if($_ENV["OC_PLUGIN_DEBUG"]) {
						error_log("exec: " . $actionType);
					}
					switch($actionType)
					{
						case "list":
							$result = $this->loadNode($actionType, $actionData);
							break;
						case "delete":
							$result = $this->delete($actionType, $actionData);
							break;
						case "tmpdownload":
							$result = $this->downloadSelectedOwncloudFilesToTmp($actionType, $actionData);
							break;
						case "save":
							$result = $this->createDirectory($actionType, $actionData);
							break;
						default:
							$this->handleUnknownActionType($actionType);
					}

				} catch (MAPIException $e) {
					if($_ENV["OC_PLUGIN_DEBUG"]) {
						error_log("mapi exception: " . $e->getMessage());
					}
				} catch (AuthException $e) {
					//TODO: rewrite this dirty code...
					if ($e->getCode() == 401) {
						$nodes["/"] = array(
							'id'	=>	"/",
							'path'	=>	"/",
							'text'	=>	$e->getCode() . " Username or Password wrong!",
							'filename'=> "error"
						);						
					} else if ($e->getCode() == 403) {
						$nodes["/"] = array(
							'id'	=>	"/",
							'path'	=>	"/",
							'text'	=>	$e->getCode() . " Could not connect to Server! Check your config!",
							'filename'=> "error"
						);
					} else if ($e->getCode() == 404) {
						$nodes["/"] = array(
							'id'	=>	"/",
							'path'	=>	"/",
							'text'	=>	$e->getCode() . " Unknown Owncloud Path or Server! Check your config!",
							'filename'=> "error"
						);
					}
					
					$response['items'] = array_values($nodes);
					$response['status']	=	true;
					$this->addActionData($actionType, $response);
					$GLOBALS["bus"]->addData($this->getResponseData());
					
					if($_ENV["OC_PLUGIN_DEBUG"]) {
						error_log("auth exception: " . $e->getMessage());
					}
				}
			}
		}
		return $result;
	}

	/**
	 * loads content of current folder - list of folders and files from Owncloud
	 * @param {Array} $actionType
	 * @param {Array} $actionData
	 */
	public function loadNode($actionType, $actionData) {	
		$nodeId = $actionData['id'];
		
		$cache_object = $this->cache->get($nodeId);
		if(!is_null($cache_object)) {
			$dir = $cache_object;
		} else {
			if (!$this->wdc->open()) {
				if($_ENV["OC_PLUGIN_DEBUG"]) {
					error_log('Error: could not open server connection');
				}
				throw new AuthException("403 Failed to connect",403);
			}
			
			/* logon to owncloud */
			$this->authorise();
			
			$dir = $this->wdc->ls($nodeId);
			$this->cache->put($nodeId, $dir);
				
			$this->wdc->close();
		}
		
		if($_ENV["OC_PLUGIN_DEBUG"]) {
			error_log("parsing dir: " . $dir);
		}
		if($dir == -401) {	// unauthorised!
			throw new AuthException("401 unauthorised",401);
		} else if($dir == -404) {	// unknown/not found!
			throw new AuthException("404 not found",404);
		}else {
			$response["item"] = $this->parseFolderContent($nodeId, $dir, $actionData);
		}
		if($_ENV["OC_PLUGIN_DEBUG"]) {
			error_log("parsing done!");
		}
		
		$response['page'] = array("start" => 0, "rowcount" => 50, "totalrowcount" => count($response["item"]));
		$response['folder'] = array("content_count" => count($response["item"]), "content_unread" => 0);
		$this->addActionData($actionType, $response);
		$GLOBALS["bus"]->addData($this->getResponseData());
		
		if($_ENV["OC_PLUGIN_DEBUG"]) {
			error_log("nodes loaded, bus data written!");
		}
	}

	/**
	 * form the structure needed for frontend
	 * for the list of folders and files
	 * @param basedir
	 * @param dir
	 * @param data
	 * @return {Array} nodes for current path folder
	 * @private
	 */
	private function parseFolderContent($basedir, $dir, $data) {
		$nodes = array();
		$oc_base = $GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/owncloud_path");
		
		
		$parentdir =  dirname($basedir);  // get parent dir
		if($parentdir != "/") {
			$parentdir = $parentdir . "/";
		}
		if($basedir != "/") {	// add a back button
			$nodes[$parentdir] = array('props' =>
				array(
					'id'	=>	urldecode($parentdir),
					'path'	=>	dirname(urldecode($parentdir)),
					'filename'=> "..",							
					'message_size'	=>	0,
					'lastmodified' => null,
					'message_class' => "IPM.Owncloud",
					'type'=> 0
				),
				'entryid' => urldecode($parentdir),
				'store_entryid' => 'owncloud',
				'parent_entryid' => dirname(urldecode($parentdir)) . "/"
			);
		}
		
		foreach($dir as $node) {
			$type = 1;
			
			$id = substr(urldecode($node['href']),strlen($oc_base));
			
			if($_ENV["OC_PLUGIN_DEBUG"]) {
				error_log("parsing: " . $id . " in base: " . $basedir);
			}
			
			$filename =  basename($id);
			$basedir2 = $basedir."/";
			if (strcmp($id,$basedir) == 0 || strcmp($id,$basedir2) == 0 ) { // current dir...
				continue;
			}
			if (strcmp($node['resourcetype'],"collection") == 0) { // we have a folder
				$type = 0;
			}
			
			$nodes[$id] = array('props' =>
				array(
					'id'	=>	urldecode($id),
					'path'	=>	dirname(urldecode($id)),
					'filename'=>$filename,							
					'message_size'	=>	$node['getcontentlength'] == null ? 0 : intval($node['getcontentlength']),
					'lastmodified' => strtotime($node['lastmodified']) * 1000,
					'message_class' => "IPM.Owncloud",
					'type'=> $type
				),
				'entryid' => urldecode($id),
				'store_entryid' => 'owncloud',
				'parent_entryid' => dirname($id) . "/"
			);
			
		}
		return array_values($nodes);
	}
	
	/**
	 * Sets access token for Owncloud's WebDav Interface
	 * @param {Array} $actionType
	 * @param {Array} $actionData
	 * @private
	 */
	private function delete($actionType, $actionData) {
		if (!$this->wdc->open()) {
			if($_ENV["OC_PLUGIN_DEBUG"]) {
				error_log('Error: could not open server connection');
			}
			throw new AuthException("403 Failed to connect",403);
		}
		
		/* logon to owncloud */
		$this->authorise();
		
		$nodeId = $actionData['entryid'];
		$http_status = $this->wdc->delete($nodeId);
		
		$this->cache->remove($nodeId); // remove cached object...
		
		$this->wdc->close();
		
		if($_ENV["OC_PLUGIN_DEBUG"]) {
			error_log("delete status: " . $http_status["status"]);
		}
		
		if(intval($http_status["status"]) > 210) {	// something went wrong...
			$msg = "Deleting item failed! (" . $http_status['status'] . ")";
			$this->sendFeedback(false, array('error' => array(
				'type' => ERROR_GENERAL,
				'info' => array(
					'original_message' => $msg,
					'display_message' => $msg
				))
			));
		} else {		
			$response['status'] = true;
			$this->addActionData($actionType, $response);
			$GLOBALS["bus"]->addData($this->getResponseData());
		}
	}
	
	/**
	 * Downloads file from the Owncloud service and saves it in tmp
	 * folder with unique name
	 * @param {Array} $actionData
	 * @private
	 */
	private function downloadSelectedOwncloudFilesToTmp($actionType, $actionData) {
		$ids = $actionData['ids'];
		$oc_base = $GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/owncloud_path");
		$response = array();
		
		if (!$this->wdc->open()) {
			if($_ENV["OC_PLUGIN_DEBUG"]) {
				error_log('Error: could not open server connection');
			}
			throw new AuthException("403 Failed to connect",403);
		}
		
		/* logon to owncloud */
		$this->authorise();
		
		$attachment_state = new AttachmentState();
		
		// DIRTY WORKAROUND TO KEEP COMPATIBILITY FOR OLDER WEBAPPS (<=1.2.1)
		if(method_exists($attachment_state,"open")) {
			$attachment_state->open();
		}
		
		foreach ($ids as $id) {
			$file = $id;
			$filename = basename($file);
			
			// DIRTY WORKAROUND TO KEEP COMPATIBILITY FOR OLDER WEBAPPS (<=1.2.1)
			if(method_exists($attachment_state,"getAttachmentTmpPath")) {
				$tmpname = $attachment_state->getAttachmentTmpPath($filename);
			} else {
				$tmpname = tempnam(TMP_PATH, stripslashes($filename));
			}
			
			$path = dirname($file);
			
			/* TODO: Improve this workaround */
			$secid = $this->randomstring();	
			$this->createSecIDFile(dirname($tmpname), $secid);
			
			if($_ENV["OC_PLUGIN_DEBUG"]) {
				error_log("Downloading: " . $filename . " to: " . $tmpname);
			}
			/* download file from webdav */
			$buffer = null;
			$http_status = $this->wdc->get($file, $buffer);
			if($_ENV["OC_PLUGIN_DEBUG"]) {
				error_log("http-status: " . $http_status);
			}			
			
			$fhandle = fopen($tmpname,'w');
			fwrite($fhandle,$buffer,strlen($buffer));
			fclose($fhandle); 
			
			$response['items'][] = array(
				'name' => $filename,
				'size' => filesize($tmpname),
				'tmpname'=>$this->getFilenameFromPath($tmpname),
				'secid' => $secid,
				'basedir' => dirname($tmpname)
			);
			
			if($_ENV["OC_PLUGIN_DEBUG"]) {
				error_log("filesize: " . filesize($tmpname));
			}
		}
		
		$this->wdc->close();
		$attachment_state->close();
		$response['status'] = true;
		$this->addActionData($actionType, $response);
		$GLOBALS["bus"]->addData($this->getResponseData());
	}
	
	/**
	 * create a new Directory
	 * @param {Array} $actionType
	 * @param {Array} $actionData
	 * @private
	 */
	private function createDirectory($actionType, $actionData) {
		if (!$this->wdc->open()) {
			if($_ENV["OC_PLUGIN_DEBUG"]) {
				error_log('Error: could not open server connection');
			}
			throw new AuthException("403 Failed to connect",403);
		}
		
		/* logon to owncloud */
		$this->authorise();
		
		$actionData = $actionData["props"];	
		$dirname = $actionData["id"];
		$http_status  = $this->wdc->mkcol($dirname); // create it !
		
		$parentdir =  dirname($dirname);  // get parent dir
		if($parentdir != "/") {
			$parentdir = $parentdir . "/";
		}
		
		$this->cache->remove($dirname);
		$this->cache->remove($parentdir);
		
		if($_ENV["OC_PLUGIN_DEBUG"]) {
			error_log("Created dir: " . $dirname . " response: " . $http_status);
		}
		
		$response = array();
		
		if(intval($http_status) > 210) {	// something went wrong...
			$msg = "Creating folder failed! (" . $http_status . ")";
			$this->sendFeedback(false, array('error' => array(
				'type' => ERROR_GENERAL,
				'info' => array(
					'original_message' => $msg,
					'display_message' => $msg
				))
			));
		} else {
			/* create the response folder object */
			$folder = array();
			
			$folder[$dirname] = array(
				'props' =>
					array(
						'id'	=>	urldecode($dirname),
						'path'	=>	dirname(urldecode($dirname)),
						'filename'=> $actionData["filename"],							
						'message_size'	=>	0,
						'lastmodified' => $actionData['lastmodified'],
						'message_class' => "IPM.Owncloud",
						'type'=> $actionData['type']
					),
				'entryid' => urldecode($dirname),
				'store_entryid' => 'owncloud',
				'parent_entryid' => $parentdir
			);
			$response['item'] = array_values($folder);
			
			$this->addActionData("update", $response);
			$GLOBALS["bus"]->addData($this->getResponseData());
		}
	}

	/**
	 * Sets access token for Owncloud's WebDav Interface
	 * @param {Array} $data
	 * @private
	 */
	private function authorise() {
		$this->wdc->set_user($GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/username"));
		$this->wdc->set_pass(base64_decode($GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/password")));
		
		if($_ENV["OC_PLUGIN_DEBUG"]) {
			error_log("auth for " . $GLOBALS["settings"]->get("zarafa/v1/contexts/owncloud/username"));
		}
	}
	
	/**
	 * @param {String} $path
	 * @return {String} filename
	 * @private
	 */
	private function getFilenameFromPath($path) {
		$pathParts = explode('/', $path);
		return end($pathParts);
	}
	
	/** 
	 * creates a random string
	 *
	 * @param {Integer} length The length of the random string
	 */
	private function randomstring($length = 6) {
		// $chars - all allowed charakters
		$chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890";

		srand((double)microtime()*1000000);
		$i = 0;
		while ($i < $length) {
			$num = rand() % strlen($chars);
			$tmp = substr($chars, $num, 1);
			$pass = $pass . $tmp;
			$i++;
		}
		return $pass;
	}
	
	/** 
	 * creates a security file, which is checked before downloading a file
	 *
	 * @param {String} secid A random id
	 */
	private function createSecIDFile($basepath, $secid) {
		$lockFile = $basepath . DIRECTORY_SEPARATOR . "secid." . $secid;
		$fh = fopen($lockFile, 'w') or die("can't open secid file");
		$stringData = date(DATE_RFC822);
		fwrite($fh, $stringData);
		fclose($fh);
	}
	
	/**
	 * Function will be used to process private items in a list response,
	 * owncloud doesn't have private items so function is overriden to not do 
	 * any processing.
	 * @param {Object} $item item properties
	 * @return {Object} item properties if its non private item otherwise empty array
	 */
	function processPrivateItem($item) {
		return $item;
	}
}
?>
